"""Helpers for handling REST-APIs."""

import functools
import urllib.parse

import requests.exceptions

from . import sessions


class Client:
    """A General purpose REST / JSON API-client."""

    def __init__(self, base_url: str, auth=None, request_timeout=2):
        """Create a new API client instance."""
        self._session = sessions.Session()
        self._session.timeout = request_timeout
        if auth:
            self._session.auth = auth
        if not base_url.endswith("/"):
            base_url += "/"
        self._base_url = base_url

    def __getattr__(self, name):
        """All undefined attributes are interpreted as HTTP verbs."""
        verb = name.upper()
        return functools.partial(self.request, verb)

    def request(self, method, *path_parts, path=None, url=None, **kwargs):
        """Send a request to the API-endpoint and handle the response.

        This accepts either of (the first of these wins):
        - A url passed as kwarg.
        - A path passed as kwarg.
        - Path parts that will be escaped and joined passed as positional arguments.
        """
        if not path and path_parts:
            path = "/".join(urllib.parse.quote_plus(part) for part in path_parts)
        if url:
            if not url.startswith(self._base_url):
                raise requests.exceptions.URLRequired(
                    f"This client only sends requests to {self._base_url}"
                )
        elif path:
            url = self._base_url + path
        response = self._session.request(method, url, **kwargs)
        response.raise_for_status()
        return response



"""Unit-tests for the auth middleware implementations."""
# pylint: disable=too-few-public-methods

from unittest import mock

import flask
import pytest
import requests_mock as rm

from .. import auth


@pytest.fixture(name="app", scope="class")
def fixture_app():
    """Define a test Flask app."""
    app = flask.Flask("test-app")
    app.config["IMPACT_STACK_AUTH_APP_URL"] = "https://auth.impact-stack.org/v1"
    app.config["IMPACT_STACK_AUTH_APP_KEY"] = "api-key"
    with app.app_context():
        yield app


@pytest.mark.usefixtures("app")
class AuthAppClientTest:
    """Test the auth-app client."""

    @staticmethod
    def test_get_token(requests_mock):
        """Test getting a token."""
        client = auth.AuthAppClient.from_app_config()
        requests_mock.post(rm.ANY, json={"token": "TOKEN.org1"})
        token = client.get_token("org1")
        assert token == "TOKEN.org1"


@pytest.mark.usefixtures("app")
class AuthAppMiddlewareTest:
    """Test the auth-app middleware."""

    @staticmethod
    def test_default_client_from_app_config():
        """Test that instantiating without a client uses client configured from the app config."""
        middleware = auth.AuthAppMiddleware("org1")
        assert isinstance(middleware.client, auth.AuthAppClient)

    @staticmethod
    def test_call_adds_header():
        """Test that the JWT token is added to the header."""
        client = mock.Mock(spec=auth.AuthAppClient)
        client.get_token.return_value = "TOKEN.org1"
        middleware = auth.AuthAppMiddleware("org1", client=client)

        request = mock.Mock(headers={})
        middleware(request)
        assert client.method_calls == [mock.call.get_token("org1")]
        assert "Authorization" in request.headers
        assert request.headers["Authorization"] == "Bearer TOKEN.org1"



import smtplib

from flask import current_app
from flask_mail import Connection as _Connection
from flask_mail import Mail as _Mail
from flask_mail import Message
from flask_mail import _Mail as __Mail

"""
Wrappers around flask_mail to support a custom MAIL FROM and local_hostname.
"""


class Connection(_Connection):
    def configure_host(self):
        kwargs = dict(local_hostname=self.mail.local_hostname)
        if self.mail.use_ssl:
            host = smtplib.SMTP_SSL(self.mail.server, self.mail.port, **kwargs)
        else:
            host = smtplib.SMTP(self.mail.server, self.mail.port, **kwargs)

        host.set_debuglevel(int(self.mail.debug))

        if self.mail.use_tls:
            host.starttls()
        if self.mail.username and self.mail.password:
            host.login(self.mail.username, self.mail.password)

        return host


class MailState(__Mail):
    def send(self, msg):
        """ Set custom headers and call parent. """
        if self.reply_to:
            msg.reply_to = self.reply_to
        with self.connect() as connection:
            connection.send(msg, self.envelope_from)

    def connect(self):
        """Opens a connection to the mail host."""
        app = getattr(self, "app", None) or current_app
        try:
            return Connection(app.extensions["mail"])
        except KeyError:
            raise RuntimeError("The curent application was not configured with Flask-Mail")


class Mail(_Mail):
    def send_message(self, *args, **kwargs):
        return self.send(Message(*args, **kwargs))

    def init_mail(self, config, debug=False, testing=False):
        state = MailState(
            config.get("MAIL_SERVER", "127.0.0.1"),
            config.get("MAIL_USERNAME"),
            config.get("MAIL_PASSWORD"),
            config.get("MAIL_PORT", 25),
            config.get("MAIL_USE_TLS", False),
            config.get("MAIL_USE_SSL", False),
            config.get("MAIL_DEFAULT_SENDER"),
            int(config.get("MAIL_DEBUG", debug)),
            config.get("MAIL_MAX_EMAILS"),
            config.get("MAIL_SUPPRESS_SEND", testing),
            config.get("MAIL_ASCII_ATTACHMENTS", False),
        )
        state.envelope_from = config.get("MAIL_ENVELOPE_FROM", None)
        state.local_hostname = config.get("MAIL_LOCAL_HOSTNAME", None)
        state.reply_to = config.get("MAIL_REPLY_TO", None)
        return state



"""Tests for logging."""
# pylint: disable=no-member,protected-access

import logging as _logging

import requests
from flask import Flask
from pythonjsonlogger.jsonlogger import JsonFormatter

from moflask import logging


def test_configuring_default_logger():
    """Test Default formatter is used."""
    app = Flask("moflask")

    logging.init_logger(app)

    assert app.logger.name == "moflask"
    assert isinstance(app.logger.handlers[-1].formatter, _logging.Formatter)


def test_configuring_file_logger():
    """Test JsonFormatter is used for log file."""
    app = Flask("moflask")
    app.config["LOG_FILE"] = "/tmp/moflask.logging_test.log"

    logging.init_logger(app)

    assert app.logger.name == "moflask"
    assert isinstance(app.logger.handlers[-1].formatter, JsonFormatter)
    assert app.logger.handlers[-1].baseFilename == "/tmp/moflask.logging_test.log"


def test_handlers_are_added_once():
    """Test that existing handlers are not added again."""
    app = Flask("moflask")
    app.config["LOG_FILE"] = "/tmp/moflask.logging_test.log"

    logging.init_logger(app)
    assert len(app.logger.handlers) == 2
    handlers1 = app.logger.handlers.copy()

    logging.init_logger(app)
    assert len(app.logger.handlers) == 2
    assert app.logger.handlers == handlers1


def test_adding_extra_filters():
    """Test extra filters are added to the handlers."""

    def extra_filter(record):
        record.test = "test"
        return True

    app = Flask("moflask")
    app.config["LOG_FILE"] = "/tmp/moflask.logging_test.log"

    logging.init_logger(app, [extra_filter])
    for handler in app.logger.handlers:
        assert extra_filter in handler.filters


def test_app_context_filter():
    """Test app context enriches the LogRecord.

    Pulls information from a Flask `currrent_app` context.
    """
    app = Flask("app_context")

    context_filter = logging.add_app_context_data
    record = _logging.makeLogRecord({})
    with app.app_context():
        context_filter(record)
    assert record.app
    assert record.app == "app_context"


def test_request_context_filter():
    """Test request context enriches the LogRecord.

    Pulls information from a Flask `request` context.
    """
    app = Flask(__name__)

    context_filter = logging.add_request_context_data
    record = _logging.makeLogRecord({})
    with app.test_request_context("/", environ_base={"REMOTE_ADDR": "127.1.2.3"}):
        context_filter(record)
    assert record.request_remote_addr
    assert record.request_remote_addr == "127.1.2.3"


def test_request_filter():
    """Test _request in extra enriches the LogRecord."""
    record = _logging.makeLogRecord({"_request": requests.Request("POST", "https://mo.flask")})
    logging.add_request_data(record)
    assert record.request_method
    assert record.request_method == "POST"
    assert record.request_url
    assert record.request_url == "https://mo.flask"


def test_response_filter():
    """Test _response in extra enriches the LogRecord."""
    response = requests.Response()
    response.status_code = 200
    # pylint: disable=protected-access
    response._content = "Hello".encode("utf-8")
    record = _logging.makeLogRecord({"_response": response})
    logging.add_response_data(record)
    assert record.response_status_code
    assert record.response_status_code == 200
    assert record.response_body
    assert record.response_body == "Hello"


def test_response_filter_with_json():
    """Test the added response body is JSON if the response was JSON."""
    response = requests.Response()
    response.status_code = 200
    # pylint: disable=protected-access
    response._content = '{"test": 1}'.encode("utf-8")
    record = _logging.makeLogRecord({"_response": response})
    logging.add_response_data(record)
    assert record.response_status_code
    assert record.response_status_code == 200
    assert record.response_body
    assert "test" in record.response_body
    assert record.response_body["test"] == 1



"""Logging with some goodies.

Provide a enhanced logging infrastructure:
- A default handler with a custom format (settings key `LOG_FORMAT`)
- A file log handler in JSON format (interesting for structured logging)
- Filters to enrich log entries with contextual information from Flask or
  Werkzeug or requests such as the request IP, the response status code,…

Settings options:
- LOG_LEVEL: log level for the logger
- LOG_FORMAT: string format for the default formatter
- LOG_DATE_FORMAT: date format for the default formatter
- LOG_FILE: path to the log file – if not set the file handler is disabled
- LOG_FILE_INCLUDE: record fields to include in the JSON output
- LOG_FILE_MAX_SIZE: maximum size of the logfile before a new file is started
- LOG_FILE_COUNT: number of old log files to keep (if 0 the max size will be exceeded)

"""

import json
import logging
import logging.handlers
import os

import flask
from pythonjsonlogger import jsonlogger

# Ensure the `current_user` variable is set for the filter below.
try:
    from flask_login import current_user
except ImportError:
    current_user = None


def init_logger(app, extra_filters=None):
    """Configure the app’s logger with custom filters and handlers."""
    logger = logging.getLogger(app.name)
    logger.setLevel(app.config.get("LOG_LEVEL", logging.DEBUG if app.debug else logging.INFO))

    # Configure a handler before accessing the logger via app.logger.
    logger.addHandler(get_default_handler(app.config, extra_filters))

    # Add a file handler too if configured.
    file_handler = get_json_file_handler(app.config, logger, extra_filters)
    if file_handler:
        logger.addHandler(file_handler)


def add_request_context_data(record):
    """Add request data from the context to the log record if available."""
    if flask.has_request_context():
        record.request_remote_addr = flask.request.remote_addr
    return True


def add_app_context_data(record):
    """Add app data from the context to the log record if available."""
    if flask.has_app_context():
        record.app = flask.current_app.name
    return True


def add_user_data(record):
    """Add user data to the log record if available."""
    if current_user:
        if current_user.is_authenticated:
            record.user = repr(current_user)
        else:
            record.user = "<AnonymousUser>"
    return True


def add_request_data(record):
    """Add contextual information about a request to the log record.

    The filter looks for the attribute `_request` in the `record`,
    tries to retrieve specific information and sets or updates attributes
    on the log record. The request is expected to be of type `requests.Request`.

    `_request` can be added by log calls in the `extra` dict.
    """
    request = getattr(record, "_request", None)
    if request:
        record.request_url = request.url
        record.request_method = request.method
    return True


def add_response_data(record):
    """Add contextual information about a response to the log record.

    The filter looks for the attribute `_response` in the `record`,
    tries to retrieve specific information and sets or updates attributes
    on the log record. The response is expected to be of type `requests.Response`.

    `_response` can be added by log calls in the `extra` dict.
    """
    response = getattr(record, "_response", None)
    if response:
        record.response_status_code = response.status_code
        try:
            record.response_body = response.json()
        except json.decoder.JSONDecodeError:
            record.response_body = response.text
    return True


def get_formatter(config):
    """Get a string formatter."""
    fmt = config.get("LOG_FORMAT", "%(asctime)s [%(levelname)s] %(name)s: %(message)s")
    datefmt = config.get("LOG_DATE_FORMAT", None)
    return logging.Formatter(fmt, datefmt)


def get_json_formatter(config):
    """Get a JSON formatter."""
    default_include = [
        "message",
        "levelname",
        "name",
        "asctime",
        "created",
        "process",
        "processName",
        "thread",
        "threadName",
        "lineno",
        "module",
        "pathname",
    ]
    include = config.get("LOG_FILE_INCLUDE", default_include)
    # Fields form the `extra` dict are automatically added by the JsonFormatter.
    return jsonlogger.JsonFormatter(" ".join(f"%({prop})" for prop in include))


def get_default_handler(config, filters=None):
    """Get a default stream handler with custom formatting."""
    handler = flask.logging.default_handler
    handler.setFormatter(get_formatter(config))
    for filter_ in filters or []:
        handler.addFilter(filter_)
    return handler


def get_json_file_handler(config, logger=None, filters=None):
    """Get a file handler with custom JSON formatting and context data."""
    log_file = config.get("LOG_FILE", os.environ.get("LOG_FILE", None))
    if log_file is None:
        return None
    # Check for existing handler before adding a new one.
    handler = next(
        filter(
            lambda h: getattr(h, "baseFilename", None) == log_file,
            logger.handlers if logger else [],
        ),
        logging.handlers.RotatingFileHandler(
            log_file,
            maxBytes=config.get("LOG_FILE_MAX_SIZE", 0),
            backupCount=config.get("LOG_FILE_COUNT", 0),
        ),
    )
    handler.setFormatter(get_json_formatter(config))
    default_filters = [
        add_app_context_data,
        add_request_context_data,
        add_request_data,
        add_response_data,
    ]
    if current_user:
        default_filters.append(add_user_data)
    for filter_ in default_filters + (filters or []):
        handler.addFilter(filter_)
    return handler



#!/usr/bin/env python

"""Package meta data for moflask."""

import setuptools

setuptools.setup(
    name="moflask",
    version="1.0.1",
    description="Re-usable flask utilities.",
    author="Roman Zimmermann",
    author_email="torotil@gmail.com",
    packages=setuptools.find_packages(),
    install_requires=[
        "Flask>=1.1",
        "python-json-logger",
    ],
    extras_require={
        "test": ["pytest", "python-json-logger"],
        "requests": ["requests"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    url="https://github.com/moreonion/moflask",
    download_url="https://github.com/moreonion/moflask/archive/refs/tags/v1.0.1.tar.gz",
)



"""Auth middlewares for the requests library."""

import urllib.parse

import flask
import requests

from . import rest


class AuthAppClient(rest.Client):
    """Client for the impact-stack auth app."""

    @classmethod
    def from_app_config(cls):
        """Create a new auth-app client instance from the app config."""
        return cls(flask.current_app.config.get("IMPACT_STACK_AUTH_APP_URL"))

    def get_token(self, organization):
        """Get a JWT token for a sub-organization."""
        api_key = flask.current_app.config.get("IMPACT_STACK_AUTH_APP_KEY")
        organization = urllib.parse.quote_plus(organization)
        response = self.post("token", organization, json=api_key)
        return response.json()["token"]


class AuthAppMiddleware:
    """Middleware for authenticating using JWT tokens.

    The middleware transparently requests an API-token from the auth-app on-demand.
    """

    def __init__(self, organization, client=None):
        """Create new auth-app requests auth middleware."""
        self.client = client or AuthAppClient.from_app_config()
        self.organization = organization

    def get_token(self):
        """Get a JWT token from the auth-app."""
        return self.client.get_token(self.organization)

    def __call__(self, request: requests.PreparedRequest):
        """Add the JWT token to the request."""
        request.headers["Authorization"] = "Bearer " + self.get_token()
        return request
